import inspect
from typing import Type, Set

from sqlalchemy import Column
from sqlalchemy.orm import class_mapper, RelationshipProperty
from sqlalchemy.orm.base import manager_of_class
from sqlalchemy.orm.decl_api import declared_attr

from sqlalchemy_orm.query import Query


class HelperBase:
    declarative_base = None

    @declared_attr
    def __tablename__(cls):
        return cls.__name__.lower()

    @classmethod
    def base(cls):
        declarative_base = cls.declarative_base
        if not declarative_base:
            return None

        mro = reversed(inspect.getmro(cls))
        for mro_cls in mro:
            if mro_cls != declarative_base and \
                    issubclass(mro_cls, declarative_base):

                if not mro_cls.__dict__.get(
                        '__abstract__',
                        False
                ):
                    return mro_cls

    @classmethod
    def is_base(cls):
        base = cls.base()
        if base:
            return cls == base
        return False

    @classmethod
    def super(cls):
        mro = inspect.getmro(cls)
        for mro_cls in mro[1:]:
            base = cls.declarative_base
            if base and issubclass(mro_cls, base):

                if not mro_cls.__dict__.get(
                        '__abstract__',
                        False
                ):
                    return mro_cls

    @classmethod
    def identifier(cls):
        return cls.__name__

    @classmethod
    def filter(cls, *args, **kwargs) -> Query:
        query = cls.query()
        if args:
            query = query.filter(*args)
        if kwargs:
            query = query.filter_by(**kwargs)

        return query

    @classmethod
    def parents(cls) -> Set[Type]:
        parents = []

        _cls = cls
        while _cls:
            _cls = _cls.super()

            if _cls:
                parents.append(_cls)

        return parents

    @classmethod
    def relations(cls, checked=None) -> Set[Type]:
        if not checked:
            checked = set()
        if hasattr(cls, '__mapper__') and hasattr(cls, '__table__'):
            relations = set()

            models = cls.models

            for foreign_key in cls.__table__.foreign_keys:
                table = foreign_key.column.table
                for model in models:
                    if hasattr(model, '__table__') and \
                            model.__table__ == table:
                        relations.add(model)
                        if model not in checked:
                            checked.add(model)
                            relations.update(model.relations(checked=checked))

            for prop in class_mapper(cls).iterate_properties:
                if isinstance(prop, RelationshipProperty):
                    model = prop.mapper.class_
                    relations.add(model)
                    if model not in checked:
                        checked.add(model)
                        relations.update(model.relations(checked=checked))

            return relations

        raise RuntimeError(
            "Attempted to retrieve relations from unmapped class '{cls}'"
        )

    @classmethod
    def is_mapped(cls):
        class_manager = manager_of_class(cls)
        return class_manager and class_manager.is_mapped

    @classmethod
    def primary_keys(cls):
        if hasattr(cls, '__mapper__'):
            return cls.__mapper__.primary_key

        _primary_keys = []
        for key, value in cls.__dict__.items():
            if isinstance(value, Column):
                if value.primary_key:
                    _primary_keys.append(value)

        if _primary_keys:
            return _primary_keys

        raise RuntimeError(
            "Attempted to retrieve primary keys from unmapped class '{cls}'"
        )

    @property
    def primary_key_values(self):
        return [getattr(self, key.name) for key in self.primary_keys()]

    def __hash__(self):
        primary_key_values = (*self.primary_key_values,)
        return hash(primary_key_values)

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __repr__(self):
        primary_key_dict = {
            key.name: getattr(self, key.name)
            for key in self.primary_keys()
        }
        return f"<{self.__class__.__name__} '{primary_key_dict}'>"

    def create(self, session):
        return session.create(self)

    def delete(self, session):
        return session.delete(self)



from sqlalchemy.orm import Query as SQLAlchemyQuery


class QueryAdapter:

    def apply(self, query, entities=None):
        return query


class Query(SQLAlchemyQuery):

    def _apply_adapters(self, *criterion):
        query = self
        criterion = [
            criteria
            for criteria in criterion
            if criteria is not None
        ]
        entities = [cd['type'] for cd in self.column_descriptions]
        for criteria in [*criterion]:
            if isinstance(criteria, QueryAdapter):
                criterion.remove(criteria)
                query = criteria.apply(query, entities)

        return criterion, query

    def filter(self, *criterion):
        criterion, query = self._apply_adapters(*criterion)
        return super(Query, query).filter(*criterion)

    def order_by(self, *criterion):
        criterion, query = self._apply_adapters(*criterion)
        return super(Query, query).order_by(*criterion)



import inspect
from typing import List, Type, Union

import sqlalchemy
import typing_inspect
from sqlalchemy import Column, ForeignKey, event
from sqlalchemy.orm import relationship
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.orm.decl_base import _add_attribute

from sqlalchemy_orm.base.helper import HelperBase
from sqlalchemy_orm.base.inheritance import InheritanceBase
from sqlalchemy_orm.typemapper import default_type_mapper


class DataclassConsistencyError(BaseException):
    pass


class MapperBase(InheritanceBase, HelperBase):
    type_mapper = default_type_mapper

    unresolved = {}
    models = {}

    @classmethod
    def __ignore_fields__(cls):
        return []

    @classmethod
    def fields(cls):
        fields = []

        def add_fields(cls_):
            ignore_fields = []

            if hasattr(cls_, "__ignore_fields__"):
                ignore_fields = cls_.__ignore_fields__()
            for key, type_ in cls_.__dict__.get('__annotations__', {}).items():
                if key not in ignore_fields:
                    default = getattr(cls_, key, None)
                    if isinstance(default, InstrumentedAttribute):
                        default = None
                    fields.append((key, type_, default))

        # fields for any mixins
        for cls_ in inspect.getmro(cls)[1:]:
            is_abstract = cls_.__dict__.get('__abstract__', False)

            if issubclass(cls_, MapperBase) and not is_abstract:
                break
            add_fields(cls_)

        # fields for current class
        add_fields(cls)

        return fields

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        fields = []

        if not cls.__inheritance__:
            for cls_ in MapperBase.models.values():
                if issubclass(cls, cls_):
                    fields.extend(cls_.fields())

        fields.extend(cls.fields())

        mappable = []

        for field in fields:

            key, type_, default = field
            flat_type = get_type(type_)

            if isinstance(flat_type, str):
                MapperBase.unresolved[(cls.identifier(), field)] = flat_type

            elif inspect.isclass(flat_type) and \
                    inspect.isclass(cls.declarative_base) and \
                    issubclass(flat_type, cls.declarative_base):
                key_ = (cls.identifier(), field)
                MapperBase.unresolved[key_] = flat_type.__name__

            else:
                mappable.append(field)

        cls.map(mappable)

        MapperBase.models[cls.__name__] = cls
        MapperBase.resolve()

        @event.listens_for(cls, 'mapper_configured')
        def receive_mapper_configured(mapper, class_):
            MapperBase.resolve()

        @event.listens_for(cls, 'init')
        def instant_defaults_listener(target, args, kwargs):
            defaults = {}
            for key, column in \
                    sqlalchemy.inspect(target.__class__).columns.items():
                if (
                        hasattr(column, 'default') and
                        column.default is not None and
                        key not in kwargs
                ):
                    if callable(column.default.arg):
                        defaults[key] = column.default.arg(target)
                    else:
                        defaults[key] = column.default.arg
            kwargs.update(defaults)

    @classmethod
    def map(cls, fields):
        base = cls.base()
        inheritance = cls.__inheritance__
        is_base = cls.is_base()
        parent = cls.super()

        if base and not is_base and inheritance:
            for pk in parent.primary_keys():
                key = pk.name
                column = pk._copy()
                fk = ForeignKey(f"{parent.__tablename__}.{key}")
                column.append_foreign_key(fk)
                setattr(cls, key, column)

        had_primary_key = False

        # Map the direct fields to columns
        for field in fields:
            key, type_, default = field

            nullable = is_nullable(type_)
            primary_key = not had_primary_key and (
                    is_base or not inheritance or not base)

            if primary_key:
                had_primary_key = True
            flat_type = get_type(type_)

            column_args = []

            column_type = cls.type_mapper.to_type(flat_type)

            if not column_type:
                raise TypeError(
                    f"Type '{flat_type}' could not be mapped to a Column Type"
                )

            column = Column(
                key,
                column_type,
                *column_args,
                primary_key=primary_key,
                nullable=nullable,
                default=default
            )

            _add_attribute(cls, key, column)

    @classmethod
    def resolve(cls):
        resolved = []

        for unresolved, dependency in MapperBase.unresolved.items():
            model, field = unresolved

            model = MapperBase.models.get(model)
            dependency = MapperBase.models.get(dependency)

            if dependency is not None and dependency.is_mapped():
                key, type_, default = field

                nullable = is_nullable(type_)

                if is_list(type_):
                    nullable = True
                    use_list = True
                    single_parent = False
                    cascade = "all, delete-orphan"
                    f_key_on = dependency
                    f_key_to = model
                    f_key_prefix = model.__tablename__

                else:
                    nullable = nullable
                    use_list = False
                    single_parent = False
                    cascade = "all, delete"
                    f_key_on = model
                    f_key_to = dependency
                    f_key_prefix = key

                f_keys = []

                for p_key in f_key_to.primary_keys():
                    f_key_name = f_key_prefix + "_" + p_key.name

                    fk = ForeignKey(
                        f"{f_key_to.__tablename__}.{p_key.name}",
                        onupdate="CASCADE",
                        ondelete="SET NULL"
                    )

                    f_key = Column(
                        f_key_name,
                        p_key.type,
                        fk
                    )

                    if hasattr(f_key_on, f_key_name):
                        existing_column = getattr(f_key_on, f_key_name)
                        validate_columns(f_key, existing_column, f_key_name)

                        f_key = existing_column

                    f_key.nullable = nullable
                    f_keys.append(f_key)
                    setattr(f_key_on, f_key_name, f_key)

                # Set the Relationship
                setattr(model, key, relationship(
                    dependency,
                    uselist=use_list,
                    single_parent=single_parent,
                    foreign_keys=f_keys,
                    cascade=cascade)
                        )
                # Mark as resolved
                resolved.append(unresolved)

        # Remove resolved dependencies
        for dependency in resolved:
            del MapperBase.unresolved[dependency]


def get_value(cls, key):
    for _cls in inspect.getmro(cls):
        if key in _cls.__dict__.keys():
            return _cls.__dict__.get(key)


def validate_columns(column_1, column_2, model_name):
    if column_1.type != column_2.type:
        if str(column_1.type) != str(column_2.type):
            raise DataclassConsistencyError(
                f"Column {model_name}.{column_1.name} "
                f"redefined with a inconsistent type"
            )

    fk_names_1 = {fk.target_fullname for fk in column_1.foreign_keys}
    fk_names_2 = {fk.target_fullname for fk in column_2.foreign_keys}

    if fk_names_1 != fk_names_2:
        raise DataclassConsistencyError(
            f"Column {model_name}.{column_1.name} "
            f"redefined with inconsistent foreign keys"
        )


def get_type(field_type) -> Union[Type, str]:
    origin = typing_inspect.get_origin(field_type)

    if origin is List or origin is list or origin is Union:
        args = typing_inspect.get_args(field_type, evaluate=True)
        none_type = type(None)
        args = [arg for arg in args if arg is not none_type]

        if len(args) != 1:
            raise TypeError(
                f"Field type '{field_type}' origin: '{origin}' "
                f"contained invalid args: '{args}'."
            )

        python_type = args[0]
    else:
        python_type = field_type

    if hasattr(python_type, '__forward_arg__'):
        python_type = python_type.__forward_arg__

    if not inspect.isclass(python_type) and not isinstance(python_type, str):
        raise TypeError(
            f"Field type '{field_type}' converted to '{python_type}'"
            f" which is an invalid Python Type"
        )

    return python_type


def is_nullable(field_type) -> bool:
    if typing_inspect.is_union_type(field_type):
        union_args = typing_inspect.get_args(field_type, evaluate=True)
        return len(union_args) == 2 and type(None) in union_args
    return False


def is_list(field_type):
    origin = typing_inspect.get_origin(field_type)
    return origin is list or origin is List



import enum

from sqlalchemy_orm.query import QueryAdapter


class OrderByDirection(enum.Enum):
    asc = "asc"
    desc = "desc"


class OrderByMultipleEntitiesException(BaseException):
    pass


try:
    from dataclasses import dataclass

except ImportError:

    class OrderByBase(QueryAdapter):

        def __init__(
                self,
                key: str,
                direction: OrderByDirection = OrderByDirection.asc
        ):
            """
            The `OrderBy` ObjectType describes an
            order_by clause for a collection.
            """
            self.key = key
            self.direction = direction

else:

    @dataclass
    class OrderByBase(QueryAdapter):
        key: str
        direction: OrderByDirection = OrderByDirection.asc


class OrderBy(OrderByBase):

    def apply(self, query, entities=None):

        if len(entities) != 1:
            raise OrderByMultipleEntitiesException(
                f"Cannot use a OrderBy when querying"
                f" multiple models: '{entities}'"
            )

        model = entities[0]

        sort_attr = getattr(model, self.key)

        if self.direction == OrderByDirection.asc:
            criterion = sort_attr.asc()
        else:
            criterion = sort_attr.desc()

        return query.order_by(criterion)



import uuid
from typing import Type

from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm import relationship
from sqlalchemy.orm.decl_api import declared_attr

from sqlalchemy_orm.base.helper import HelperBase
from sqlalchemy_orm.patterns.taggable import Taggable
from sqlalchemy_orm.utils import Memoized


class TagBase(HelperBase):
    tag: str

    __inheritance__ = False

    @classmethod
    @Memoized
    def mixin(cls) -> Type[Taggable]:
        tag_cls: Type[TagBase] = cls

        class Mixin(Taggable, HelperBase):

            @classmethod
            def relations(cls, checked=None):
                super_relations = super().relations(checked=checked)
                return super_relations & {cls.tag_association_cls()}

            @classmethod
            @Memoized
            def tag_association_cls(cls):
                class TagAssociation(tag_cls.declarative_base):
                    id: str = uuid.uuid4
                    tag: tag_cls
                    parent: cls = None

                    __tablename__ = cls.__tablename__ + '_tag'
                    __inheritance__ = False

                    @declared_attr
                    def tag_str(self):
                        return association_proxy('tag', 'tag')

                tag_name = TagAssociation.__name__
                tag_module = TagAssociation.__module__

                TagAssociation.__name__ = cls.__name__ + tag_name
                TagAssociation.__module__ = cls.__module__ + tag_module

                return TagAssociation

            @declared_attr
            def tag_associations(self):
                return relationship(
                    self.tag_association_cls(),
                    cascade="all, delete, delete-orphan"
                )

            @declared_attr
            def tags(self):
                TagAssociation = self.tag_association_cls()
                Tag: Type[TagBase] = tag_cls

                def creator(tag):
                    tag_obj = Tag(tag=tag)
                    return TagAssociation(tag=tag_obj)

                return association_proxy(
                    'tag_associations',
                    'tag_str',
                    creator=creator
                )

        return Mixin



# flake8: noqa

from sqlalchemy_orm.base.base import Model
from sqlalchemy_orm.custom_types import EnumType, UUIDType
from sqlalchemy_orm.database import Database
from sqlalchemy_orm.typemapper import TypeMapper



from typing import Callable, Type

from sqlalchemy.orm.decl_api import declarative_base, DeclarativeMeta

from sqlalchemy_orm.base.helper import HelperBase
from sqlalchemy_orm.base.inheritance import InheritanceBase
from sqlalchemy_orm.base.mapper import MapperBase
from sqlalchemy_orm.typemapper import TypeMapper


class DataRMBase(MapperBase, InheritanceBase, HelperBase):
    pass


class BaseDeclarativeMeta(DeclarativeMeta):
    def __init__(cls, classname, bases, dict_, **kw):
        dict_.update(cls.__dict__)
        super().__init__(classname, bases, dict_, **kw)


def base_factory(type_map=None, base=None):
    if base and type_map is not None:
        base.type_mapper = type_map

    if base:
        combined_base = type("Base", (base, DataRMBase), {})
    else:
        combined_base = DataRMBase

    if base is None and type_map is not None:
        combined_base.type_mapper = type_map

    model: DataRMBase = declarative_base(
        cls=combined_base,
        name="Base",
        metaclass=BaseDeclarativeMeta
    )
    model.declarative_base = model

    return model


Model: Callable[[TypeMapper, Type], Type] = base_factory



from typing import List

from sqlalchemy import or_, and_

from sqlalchemy_orm.query import QueryAdapter

try:
    from dataclasses import dataclass

except ImportError:

    class FilterBase(QueryAdapter):

        # noinspection PyPep8Naming
        def __init__(
                self,
                or_: List['Filter'] = None,
                and_: List['Filter'] = None
        ):
            self.or_ = or_
            self.and_ = and_
else:

    @dataclass
    class FilterBase(QueryAdapter):
        and_: List['Filter'] = None
        or_: List['Filter'] = None


class Filter(FilterBase):

    def apply(self, query, entities=None):
        return query.filter(*self.criterion(entities=entities))

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def criterion(self, entities=None):
        if self.or_ and self.and_:
            raise ValueError(
                "Filter applied with both logical operators 'OR' and 'AND'"
            )

        criterion = []

        if self.or_:
            criterion = [
                or_(*[
                    and_(*filter.criterion(entities=entities))
                    for filter in self.or_
                ])
            ]

        if self.and_:
            criterion = [
                and_(*[
                    and_(*filter.criterion(entities=entities))
                    for filter in self.and_
                ])
            ]

        return criterion



import enum
import uuid
from typing import Type, Optional

from sqlalchemy import Enum as SQLAlchemyEnum
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.sql.type_api import TypeEngine
from sqlalchemy.types import TypeDecorator, CHAR

from sqlalchemy_orm.typemapper import TypeFactory


class UUIDType(TypeDecorator):
    """Platform-independent GUID type.

    Uses PostgreSQL's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.

    """
    impl = CHAR
    cache_ok = True

    @property
    def python_type(self):
        return uuid.UUID

    def load_dialect_impl(self, dialect):
        if dialect.name == 'postgresql':
            return dialect.type_descriptor(UUID())
        else:
            return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == 'postgresql':
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                return "%.32x" % uuid.UUID(value).int
            else:
                # hexstring
                return "%.32x" % value.int

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        else:
            if not isinstance(value, uuid.UUID):
                value = uuid.UUID(value)
            return value


class EnumType(TypeFactory):

    def to_type(self, python_type: Type) -> Optional[Type[TypeEngine]]:
        if issubclass(python_type, enum.Enum):
            class Enum(SQLAlchemyEnum):
                _python_type = python_type

                def __init__(self, *enums, **kw):
                    super().__init__(Enum._python_type, *enums, **kw)

            return Enum



from sqlalchemy import Column, String
from sqlalchemy.orm.decl_api import declared_attr

from sqlalchemy_orm.base.helper import HelperBase


class InheritanceBase(HelperBase):
    __inheritance__ = True

    @declared_attr
    def _type(cls):
        if cls.__inheritance__:
            return Column(String)

    @declared_attr
    def __mapper_args__(cls):
        mapper_args = {}

        if cls.__inheritance__:
            mapper_args["polymorphic_identity"] = cls.__name__

            if cls.is_base():
                mapper_args["polymorphic_on"] = cls._type

        else:
            mapper_args["concrete"] = True

        return mapper_args



import uuid

from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm import relationship
from sqlalchemy.orm.decl_api import declared_attr

from sqlalchemy_orm.base.helper import HelperBase
from sqlalchemy_orm.patterns.taggable import Taggable
from sqlalchemy_orm.utils import Memoized


class TagMixin(Taggable):

    @classmethod
    @Memoized
    def tag_cls(cls):
        if not issubclass(cls, HelperBase):
            raise TypeError(
                f"TagMixin class {cls} must be a subclass of {HelperBase}."
            )

        class Tag(cls.declarative_base):
            id: str = uuid.uuid4
            tag: str
            parent: cls = None

            __tablename__ = cls.__tablename__ + '_tag'
            __inheritance__ = False

        Tag.__name__ = cls.__name__ + Tag.__name__
        Tag.__module__ = cls.__module__ + Tag.__module__

        return Tag

    @declared_attr
    def tags_relationship(self):
        Tag = self.tag_cls()
        return relationship(Tag, cascade="all, delete, delete-orphan")

    @declared_attr
    def tags(self):
        return association_proxy(
            'tags_relationship',
            'tag',
            creator=lambda tag: self.tag_cls()(tag=tag)
        )



from typing import Iterable

from sqlalchemy import create_engine, inspect, MetaData
from sqlalchemy.orm import sessionmaker, RelationshipProperty
from sqlalchemy.orm.clsregistry import _ModuleMarker

from sqlalchemy_orm.base.helper import HelperBase
from sqlalchemy_orm.query import Query
from sqlalchemy_orm.session import Session
from sqlalchemy_orm.utils import find_subclasses


class DatabaseException(BaseException):
    pass


class Database:

    def __init__(self, url, session_cls=Session, query_cls=Query):
        self.url = url
        self.engine = create_engine(url)
        self.session_maker = sessionmaker(
            class_=session_cls,
            bind=self.engine,
            query_cls=query_cls
        )
        self.declarative_base = None

    def entity_relationship_diagram(self, embed_html=True) -> str:
        """
        ERD in dot format
        """

        from sqlalchemy_orm.eralchemy.main import \
            _intermediary_to_dot, \
            all_to_intermediary

        tables, relationships = all_to_intermediary(
            self.declarative_base.metadata,
            schema="Schema"
        )
        output = _intermediary_to_dot(tables, relationships)

        if embed_html:
            output = """
            <!DOCTYPE html>
            <meta charset="utf-8">
            <body>
            <script src="https://d3js.org/d3.v5.min.js"></script>
            <script src="https://unpkg.com/@hpcc-js/wasm@0.3.11/dist/index
            .min.js"></script>
            <script src="https://unpkg.com/d3-graphviz@3.0.5/build/d3
            -graphviz.js"></script>
            <div id="graph" style="text-align: center;"></div>
            <script>
            d3.select("#graph").graphviz()
            .renderDot('DOT_DATA');
            </script>
            """.replace("DOT_DATA", output).replace('\n', '').replace('\r', '')

        return output

    def session(self) -> Session:
        return self.session_maker()

    def create_all(self, base):
        base.metadata.create_all(bind=self.engine, checkfirst=True)
        self.declarative_base = base

    def create(
            self,
            *models,
            create_relations=True,
            create_polymorphic_parents=True
    ):
        if create_relations:
            models_with_relations = {*models}

            for model in models:
                if issubclass(model, HelperBase):
                    models_with_relations.update(model.relations())
                    self.declarative_base = model.declarative_base

            models = models_with_relations

        if create_polymorphic_parents:
            parents = set()

            for model in models:
                if issubclass(model, HelperBase):
                    parents.update(model.parents())

            models.update(parents)

        get_metadata(models).create_all(
            bind=self.engine,
            tables=[model.__table__ for model in models],
            checkfirst=True
        )

    def delete(self, *models):
        get_metadata(models).drop_all(
            bind=self.engine,
            tables=[model.__table__ for model in models],
            checkfirst=True
        )

    def wipe(self):
        meta = MetaData()
        meta.reflect(bind=self.engine)
        meta.drop_all(
            bind=self.engine,
            tables=meta.sorted_tables,
            checkfirst=True
        )

        if not self.is_empty():
            raise DatabaseException(
                f"Database '{self}' was not empty after wipe."
            )

    def is_empty(self):
        table_names = inspect(self.engine).get_table_names()
        return not table_names

    def validate(self, base, logger=None):
        """Check whether the current db matches the models

        Check that all tables exist with all columns

        What is not checked:

        * Column types are not verified
        * Relationships are not verified

        :return: True if all declared Models and
        Attributes have corresponding tables and columns.
        """

        iengine = inspect(self.engine)

        errors = False

        tables = iengine.get_table_names()

        # Go through all Model subclasses
        for model_class in find_subclasses(base):
            module_marker = isinstance(model_class, _ModuleMarker)

            is_abstract = model_class.__dict__.get('__abstract__', False)

            if module_marker or is_abstract:
                continue

            table = model_class.__tablename__
            if table in tables:
                # Check all columns are found

                columns = [c["name"] for c in iengine.get_columns(table)]
                mapper = inspect(model_class)

                for column_prop in mapper.attrs:
                    if isinstance(column_prop, RelationshipProperty):
                        # TODO: Add checks for relations
                        pass
                    else:
                        for column in column_prop.columns:
                            # Assume normal flat column
                            if column.key not in columns:
                                if logger:
                                    logger.info(
                                        f"Column '{table}.{column.key}' does"
                                        f" not exist in db '{self.engine}'"
                                    )
                                errors = True
            else:
                if logger:
                    logger.info(
                        "Table '{table}' does not exist in db '{self.engine}'"
                    )

                errors = True

        return not errors

    def __contains__(self, model):
        connect_ = self.engine.connect()
        return self.engine.dialect.has_table(connect_, model.__tablename__)

    def __repr__(self):
        return f"<Database url='{self.url}'>"


def get_metadata(models: Iterable[object]):
    if not models:
        raise ValueError("Metadata must have at least one model specified")

    metadata = None

    for model in models:
        if hasattr(model, 'metadata'):
            if metadata is None:
                metadata = model.metadata

            if model.metadata != metadata:
                raise ValueError("Metadata mismatch within list of models")
        else:
            raise ValueError("All models must specify a 'metadata' attribute")

    return metadata



import io

from setuptools import setup, find_packages

with io.open('README.md', 'rt', encoding='utf8') as f:
    readme = f.read()

with io.open('VERSION') as version_file:
    version = version_file.read().strip().lower()
    if version.startswith("v"):
        version = version[1:]

setup(
    name='sqlalchemy-orm',
    version=version,
    license='MIT',
    packages=find_packages(),
    include_package_data=True,
    author='Robert Parker',
    author_email='rob@parob.com',
    url='https://gitlab.com/parob/sqlalchemy-orm',
    download_url=f'https://gitlab.com/parob/sqlalchemy-orm/-/archive/v{version}/datarm-v{version}.tar.gz',
    keywords=['SQLAlchemy', 'ORM'],
    description='Data Relation Mapping framework for Python.',
    long_description=readme,
    long_description_content_type='text/markdown',
    install_requires=[
        'SQLAlchemy',
        'inflection',
        'cached-property',
        'typing_inspect'
    ],
    extras_require={
        'dev': [
            'pytest',
            'pytest-cov',
            'coverage',
            'faker'
        ]
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)




from typing import Type, Optional, List

from sqlalchemy.sql.type_api import TypeEngine

from sqlalchemy_orm.utils import get_all_subclasses


class TypeFactory:

    def __init__(self, type_map=None):
        pass

    def to_type(self, python_type: Type) -> Optional[Type[TypeEngine]]:
        pass


class TypeMapper:

    def __init__(self, type_map=None, types=None):
        if not types:
            types = []

        if not type_map:
            type_map = {}

        self.type_map = {}
        self.type_factories: List[TypeFactory] = []

        for type_ in types + get_all_subclasses(TypeEngine):
            if issubclass(type_, TypeFactory):
                self.type_factories.append(type_(type_map=self))
                continue

            try:
                if callable(type_):
                    type_obj = type_()
                else:
                    type_obj = type_

                if not hasattr(type_obj, '__visit_name__'):
                    continue

                python_type = type_obj.python_type

            except (NotImplementedError, AssertionError, TypeError):
                if type_ in types:
                    raise TypeError(
                        f"Custom type '{type}' did not specify a Python type"
                    )
            else:
                type_map = self.type_map
                if python_type not in type_map.keys() or \
                        issubclass(type_map[python_type].__class__, type_):
                    self.type_map[python_type] = type_obj
                elif type_obj in self.type_map.values():
                    raise TypeError(
                        f"Types must be consistently mapped in both "
                        f"directions, '{type}' already mapped."
                    )

        self.type_map.update(type_map)

    def to_type(self, python_type: Type) -> Optional[Type[TypeEngine]]:
        column_type = self.type_map.get(python_type)

        if column_type is not None:
            return column_type

        for type_factory in self.type_factories:
            column_type = type_factory.to_type(python_type=python_type)
            if column_type:
                self.type_map[python_type] = column_type
                return column_type


default_type_mapper = TypeMapper()



from sqlalchemy.orm import Session as SQLAlchemySession


class Session(SQLAlchemySession):

    def create(self, *instances):
        self.add_all(instances)

    def delete(self, *instances):
        for instance in instances:
            super().delete(instance)
